import logging

from autogluon.core.constants import BINARY, MULTICLASS, REGRESSION, SOFTCLASS

logger = logging.getLogger(__name__)


# TODO: Add weight support?
# TODO: Can these be optimized? What computational cost do they have compared to the default catboost versions?
class CustomMetric:
    def __init__(self, metric, is_higher_better, needs_pred_proba):
        self.metric = metric
        self.is_higher_better = is_higher_better
        self.needs_pred_proba = needs_pred_proba

    @staticmethod
    def get_final_error(error, weight):
        return error

    def is_max_optimal(self):
        return self.is_higher_better

    def evaluate(self, approxes, target, weight):
        raise NotImplementedError


def get_catboost_metric_from_ag_metric(metric, problem_type):
    if problem_type == SOFTCLASS:
        from .catboost_softclass_utils import SoftclassCustomMetric
        if metric.name != 'soft_log_loss':
            logger.warning("Setting metric=soft_log_loss, the only metric supported for softclass problem_type")
        return SoftclassCustomMetric(metric=None, is_higher_better=True, needs_pred_proba=True)
    elif problem_type == BINARY:
        metric_map = dict(
            log_loss='Logloss',
            accuracy='Accuracy',
            roc_auc='AUC',
            f1='Logloss',  # f1 uses Logloss because f1 in CatBoost is not reliable (causes errors between versions)
            f1_macro='Logloss',
            f1_micro='Logloss',
            f1_weighted='Logloss',
            balanced_accuracy='BalancedAccuracy',
            recall='Recall',
            recall_macro='Recall',
            recall_micro='Recall',
            recall_weighted='Recall',
            precision='Precision',
            precision_macro='Precision',
            precision_micro='Precision',
            precision_weighted='Precision',
        )
        metric_class = metric_map.get(metric.name, 'Logloss')
    elif problem_type == MULTICLASS:
        metric_map = dict(
            log_loss='MultiClass',
            accuracy='Accuracy',
        )
        metric_class = metric_map.get(metric.name, 'MultiClass')
    elif problem_type == REGRESSION:
        metric_map = dict(
            mean_squared_error='RMSE',
            root_mean_squared_error='RMSE',
            mean_absolute_error='MAE',
            median_absolute_error='MedianAbsoluteError',
            r2='R2',
        )
        metric_class = metric_map.get(metric.name, 'RMSE')
    else:
        raise AssertionError(f'CatBoost does not support {problem_type} problem type.')

    return metric_class



import numpy as np
from collections import OrderedDict
from scipy.sparse import hstack, csr_matrix
from sklearn.base import BaseEstimator, TransformerMixin
from autogluon.core.constants import BINARY, MULTICLASS, REGRESSION, SOFTCLASS

from ..tabular_nn.utils.categorical_encoders import OneHotMergeRaresHandleUnknownEncoder


_ag_to_xgbm_metric_dict = {
    BINARY: dict(
        accuracy='error',
        log_loss='logloss',
        roc_auc='auc'
    ),
    MULTICLASS: dict(
        accuracy='merror',
        log_loss='mlogloss',        
    ),
    REGRESSION: dict(
        mean_absolute_error='mae',
        mean_squared_error='rmse', # TODO: not supported from default eavl metric. Firstly, use `rsme` refenced by catboost model.
        root_mean_squared_error='rmse',

    ),
}


def convert_ag_metric_to_xgbm(ag_metric_name, problem_type):
    return _ag_to_xgbm_metric_dict.get(problem_type, dict()).get(ag_metric_name, None)


def _softmax(x):
    return np.exp(x) / np.sum(np.exp(x), axis=1, keepdims=True)


def _sigmoid(x):
    return 1.0 / (1.0 + np.exp(-x))


def func_generator(metric, is_higher_better, needs_pred_proba, problem_type):
    if needs_pred_proba:
        if problem_type == MULTICLASS:
            def function_template(y_hat, data):
                y_true = data.get_label()
                y_hat = _softmax(y_hat)
                res = metric(y_true, y_hat)
                return metric.name, -1 * res if is_higher_better else res
        elif problem_type == SOFTCLASS:
            def function_template(y_hat, data):
                y_true = data.get_label()
                y_hat = _softmax(y_hat)
                res = metric(y_true, y_hat)
                return metric.name, -1 * res if is_higher_better else res
        elif problem_type == BINARY:
            def function_template(y_hat, data):
                y_true = data.get_label()
                y_hat = _sigmoid(y_hat)
                res = metric(y_true, y_hat)
                return metric.name, -1 * res if is_higher_better else res
        else:
            def function_template(y_hat, data):
                y_true = data.get_label()
                res = metric(y_true, y_hat)
                return metric.name, -1 * res if is_higher_better else res
    else:
        if problem_type == MULTICLASS:
            def function_template(y_hat, data):
                y_true = data.get_label()
                y_hat = y_hat.argmax(axis=1)
                res = metric(y_true, y_hat)
                return metric.name, -1 * res if is_higher_better else res
        elif problem_type == SOFTCLASS:
            def function_template(y_hat, data):
                y_true = data.get_label()
                y_hat = y_hat.argmax(axis=1)
                res = metric(y_true, y_hat)
                return metric.name, -1 * res if is_higher_better else res
        elif problem_type == BINARY:
            def function_template(y_hat, data):
                y_true = data.get_label()
                y_hat = _sigmoid(y_hat)
                y_hat = np.round(y_hat)
                res = metric(y_true, y_hat)
                return metric.name, -1 * res if is_higher_better else res
        else:
            def function_template(y_hat, data):
                y_true = data.get_label()
                res = metric(y_true, y_hat)
                return metric.name, -1 * res if is_higher_better else res

    return function_template


class OheFeatureGenerator(BaseEstimator, TransformerMixin):
    def __init__(self, max_levels=None):
        self._feature_map = OrderedDict()  # key: feature_name, value: feature_type
        self.labels = OrderedDict()
        self.cat_cols = []
        self.other_cols = []
        self.ohe_encs = None
        self.max_levels=max_levels

    def fit(self, X, y=None):
        self.cat_cols = list(X.select_dtypes(include='category').columns)
        self.other_cols = list(X.select_dtypes(exclude='category').columns)
        self.ohe_encs = OneHotMergeRaresHandleUnknownEncoder(max_levels=self.max_levels)

        if self.cat_cols:
            self.ohe_encs.fit(X[self.cat_cols])
            assert len(self.cat_cols) == len(self.ohe_encs.categories_)
            
            for cat_col, categories in zip(self.cat_cols, self.ohe_encs.categories_):
                categories_ = categories.tolist()
                self.labels[cat_col] = categories_
                # Update feature map ({name: type})
                for category in categories_:
                    self._feature_map[f"{cat_col}_{category}"] = 'i'  # one-hot encoding data type is boolean

        if self.other_cols:
            for c in self.other_cols:
                self._feature_map[c] = 'int' if X[c].dtypes == int else 'float'
        return self

    def transform(self, X, y=None):
        X_list = []
        if self.cat_cols:
            X_list.append(self.ohe_encs.transform(X[self.cat_cols]))
        if self.other_cols:
            X_list.append(csr_matrix(X[self.other_cols]))
        return hstack(X_list, format="csr")

    def get_feature_names(self):
        return list(self._feature_map.keys())

    def get_feature_types(self):
        return list(self._feature_map.values())

    def get_original_feature_names(self):
        return self.cat_cols + self.other_cols



import logging

from autogluon.core.constants import BINARY

IGNORE = 'ignore'
ONLY = 'only'
INCLUDE = 'include'

logger = logging.getLogger(__name__)

preprocess_params_set = {
    'vectorizer_dict_size',
    'proc.ngram_range',
    'proc.skew_threshold',
    'proc.impute_strategy',
    'penalty',
    'handle_text',
}


def get_param_baseline():
    default_params = {
        'C': 1,
        'vectorizer_dict_size': 75000,  # size of TFIDF vectorizer dictionary; used only in text model
        'proc.ngram_range': (1, 5),  # range of n-grams for TFIDF vectorizer dictionary; used only in text model
        'proc.skew_threshold': 0.99,  # numerical features whose absolute skewness is greater than this receive special power-transform preprocessing. Choose big value to avoid using power-transforms
        'proc.impute_strategy': 'median',  # strategy argument of sklearn.SimpleImputer() used to impute missing numeric values
        'penalty': 'L2',  # regularization to use with regression models
        'handle_text': IGNORE, # how text should be handled: `ignore` - don't use NLP features; `only` - only use NLP features; `include` - use both regular and NLP features
    }
    return default_params


def _get_solver(problem_type):
    if problem_type == BINARY:
        # TODO explore using liblinear for smaller datasets
        solver = 'lbfgs'
    else:
        solver = 'lbfgs'
    return solver



# Code in this script based on work by Milan Cvitkovic,
# Xin Huang, Ashish Khetan and Zohar Karnin.

import copy

import torch
import torch.nn as nn

from .tab_model_base import TabModelBase


class TabTransformer(TabModelBase):
    """
    Transformer model for tabular data, can also be used for semi-supervised learning.
    This is the internal transformer model embedding that will have further fully connected layers (TabNet) to
    actually produce predictions.
    """

    def __init__(self, n_cont_embeddings, n_layers, n_heads, hidden_dim, tab_readout, column_embedding, orig_emb_resid,
                 fix_attention, n_shared_embs=8, shared_embedding_added=False, **kwargs):
        super().__init__(n_cont_features=kwargs['n_cont_features'], norm_class_name=kwargs['norm_class_name'],
                         cat_feat_origin_cards=kwargs['cat_feat_origin_cards'], max_emb_dim=kwargs['max_emb_dim'],
                         p_dropout=kwargs['p_dropout'], one_hot_embeddings=kwargs['one_hot_embeddings'],
                         drop_whole_embeddings=kwargs['drop_whole_embeddings'])

        from .modified_transformer import TransformerEncoderLayerModified

        self.n_cont_embeddings = n_cont_embeddings
        self.hidden_dim = hidden_dim
        self.readout = tab_readout
        self.orig_emb_resid = orig_emb_resid

        # Overwriting some TabModelBase options

        self.cat_initializers = nn.ModuleDict()
        if isinstance(self.cat_feat_origin_cards, list):
            self.n_embeddings = len(self.cat_feat_origin_cards) + (n_cont_embeddings if self.n_cont_features else 0)
        else:
            self.n_embeddings = None

        self.cat_initializers = nn.ModuleDict()

        from .tab_transformer_encoder import EmbeddingInitializer
        for col_name, card in self.cat_feat_origin_cards:
            self.cat_initializers[col_name] = EmbeddingInitializer(
                num_embeddings=card,
                max_emb_dim=self.max_emb_dim,
                p_dropout=self.p_dropout,
                minimize_emb_dim=False,
                drop_whole_embeddings=self.drop_whole_embeddings,
                one_hot=False,
                out_dim=self.hidden_dim,
                shared_embedding=column_embedding,
                n_shared_embs=n_shared_embs,
                shared_embedding_added=shared_embedding_added,
            )
        if self.n_cont_features:
            self.cont_norm = self.get_norm(self.n_cont_features)
            self.cont_initializer = nn.Linear(self.n_cont_features, hidden_dim * n_cont_embeddings)
            self.cont_init_norm = self.get_norm(hidden_dim * n_cont_embeddings)

        if self.readout == 'readout_emb':
            self.readout_emb = nn.Parameter(
                torch.zeros(1, hidden_dim).uniform_(-1, 1))  # We do the readout from a learned embedding
            self.n_embeddings += 1

        if fix_attention is True:
            self.n_cat_embeddings = len(self.cat_feat_origin_cards)
            self.tfmr_layers = nn.ModuleList([TransformerEncoderLayerModified(d_model=hidden_dim,
                                                                               n_cat_embeddings=self.n_cat_embeddings,
                                                                               nhead=n_heads,
                                                                               dim_feedforward=4 * hidden_dim,
                                                                               dropout=self.p_dropout,
                                                                               activation='gelu') for _ in
                                              range(n_layers)])
        else:
            self.tfmr_layers = nn.ModuleList([nn.TransformerEncoderLayer(d_model=hidden_dim,
                                                                         nhead=n_heads,
                                                                         dim_feedforward=4 * hidden_dim,
                                                                         dropout=self.p_dropout,
                                                                         activation='gelu') for _ in
                                              range(n_layers)])

    def init_input(self, input):
        feats = [init(input[:, i]) for i, init in enumerate(self.cat_initializers.values())]

        if self.readout == 'readout_emb':
            readout_emb = self.readout_emb.expand_as(feats[0])
            feat_embs = torch.stack([readout_emb] + feats,
                                    dim=0)  # (n_feat_embeddings + 1) x batch x hidden_dim
        else:
            feat_embs = torch.stack(feats, dim=0)  # n_feat_embeddings x batch x hidden_dim
        return feat_embs

    def run_tfmr(self, feat_embs):
        orig_feat_embs = feat_embs
        all_feat_embs = [feat_embs]
        for layer in self.tfmr_layers:
            feat_embs = layer(feat_embs)
            all_feat_embs.append(feat_embs)
            if self.orig_emb_resid:
                feat_embs = feat_embs + orig_feat_embs

        if self.readout == 'readout_emb':
            out = self.fc_out(feat_embs[0])
        elif self.readout == 'mean':
            out = torch.mean(feat_embs, dim=0)
            out = self.fc_out(out)
        elif self.readout == 'concat_pool':
            all_feat_embs = torch.cat(all_feat_embs, dim=0)

            max = all_feat_embs.max(dim=0).values
            mean = all_feat_embs.mean(dim=0)
            last_layer = feat_embs.transpose(0, 1).reshape(feat_embs.shape[1], -1)
            out = torch.cat((last_layer, max, mean), dim=1)
            out = self.fc_out(out)
        elif self.readout == 'concat_pool_all':
            feat_embs_all_layers = []
            for each_feat_embs in [all_feat_embs[0], all_feat_embs[-1]]:
                feat_embs_all_layers.append(each_feat_embs.transpose(0, 1).reshape(each_feat_embs.shape[1], -1))

            all_feat_embs = torch.cat(all_feat_embs, dim=0)
            max = all_feat_embs.max(dim=0).values
            mean = all_feat_embs.mean(dim=0)

            feat_embs_all_layers.append(max)
            feat_embs_all_layers.append(mean)
            out = torch.cat(feat_embs_all_layers, dim=1)
            out = self.fc_out(out)
        elif self.readout == 'concat_pool_add':
            orig_feat_embs_cp = copy.deepcopy(orig_feat_embs.detach())
            # ce_dim = orig_feat_embs_cp.shape[-1]//8
            # orig_feat_embs_cp[:, :, ce_dim:] = 0

            last_layer = feat_embs.transpose(0, 1).reshape(feat_embs.shape[1], -1)
            last_layer += orig_feat_embs_cp.transpose(0, 1).reshape(orig_feat_embs_cp.shape[1], -1)

            all_feat_embs = torch.cat(all_feat_embs, dim=0)
            max = all_feat_embs.max(dim=0).values
            mean = all_feat_embs.mean(dim=0)

            out = torch.cat([last_layer, max, mean], dim=1)

        elif self.readout == 'all_feat_embs':
            out = feat_embs

        elif self.readout == 'mean_feat_embs':
            out = feat_embs.mean(dim=0)

        elif self.readout == 'none':
            out = feat_embs.transpose(1, 0)

        return out

    def forward(self, input):
        """
        Returns logits for output classes
        """
        feat_embs = self.init_input(input)
        out = self.run_tfmr(feat_embs)
        return out
